/*
 * ltc4266-driver.c
 *
 * LTC4266 driver
 *
 * Author: Cradlepoint Technology, Inc.  <source@cradlepoint.com>
 *         Kyle Swenson <kswenson@cradlepoint.com>
 *
 * Copyright 2019 Cradlepoint Technology, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#include <linux/kernel.h>

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/idr.h>
#include <linux/device.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>


#include "../pse-port.h"
#include "../pse-class.h"

#include "ltc4266.h"
#include "ltc4266-regdefs.h"
#define DRV_VERSION "16Jun2017"


#define LTC4266_INTERRUPT_SOURCES	(LTC4266_INT_SUPPLY|LTC4266_INT_TSTART|LTC4266_INT_TCUT|LTC4266_INT_CLASS|LTC4266_INT_DET|LTC4266_INT_DIS|LTC4266_INT_PWRENA|LTC4266_INT_PWRGD)
#define LTC4266_INTERRUPT_PIN		18

static struct i2c_driver ltc4266_driver;

#define to_ppd(dev)  to_pse_port_device(dev)

/* Table for current limiting */
#define ILIM_ENTRIES 19
static uint16_t ltc4266_ilim_table[ILIM_ENTRIES][2] = {
				{ 0x88, 106 },
				{ 0x08, 213 },
				{ 0x89, 319 },
				{ 0x80, 425 },
				{ 0x8A, 531 },
				{ 0x90, 638 },
				{ 0x9A, 744 },
				{ 0xC0, 850 },
				{ 0xCA, 956 },
				{ 0xD0, 1063 },
				{ 0xDA, 1169 },
				{ 0xE0, 1275 },
				{ 0x49, 1488 },
				{ 0x40, 1700 },
				{ 0x4A, 1913 },
				{ 0x50, 2125 },
				{ 0x5A, 2338 },
				{ 0x60, 2550 },
				{ 0x52, 2975 } };


static inline int ilim_from_reg_val(uint8_t regval)
{
	int i = 0;

	for (i = 0; i <  ILIM_ENTRIES; i++) {
		if (ltc4266_ilim_table[i][0] == regval)
			return ltc4266_ilim_table[i][1];
	}
	return 0; /* There is nothing to correctly return here */
}
static inline uint8_t ilim_to_reg_val(int ilim)
{
	int i = 0;

	for (i = 1; i < ILIM_ENTRIES; i++) {
		if (ltc4266_ilim_table[i][1] > ilim)
			return ltc4266_ilim_table[i-1][0];
	}

	if (ilim > 425)
		return 0xC0;

	/*Failing, that, we're going to return the chip's default of 425 mA*/
	return 0x80;
}

/* class_map: Map LTC4266- specific classification values to the PSE port class
 * classification values
*/
uint8_t class_map(uint8_t clsstat)
{
	clsstat &= 0x07;
	switch (clsstat) {
	case LTC4266_CLASS_1:
		return CLASS_1;
	case LTC4266_CLASS_2:
		return CLASS_2;
	case LTC4266_CLASS_3:
		return CLASS_3;
	case LTC4266_CLASS_4:
		return CLASS_4;
	case LTC4266_CLASS_0:
		return CLASS_UNKNOWN;
	case LTC4266_CLASS_UNKNOWN:
		/*fall*/
	case LTC4266_CLASS_ERR:
		/*through*/
	case LTC4266_CLASS_OVERCURRENT:
		/*to */
	default:
		/*here*/
		return CLASS_INVALID;
	}
}

uint8_t detect_map(uint8_t detstat)
{
	detstat &= 0x07;
	switch (detstat) {
	case LTC4266_DETECT_INCOMPLETE:
		return DETECT_ERROR;
	case LTC4266_DETECT_RSIG_LOW:
	case LTC4266_DETECT_SHORT:
		return DETECT_NORMAL_DEVICE;
	case LTC4266_DETECT_HIGHCAP:
		return DETECT_LEGACY_DEVICE;
	case LTC4266_DETECT_RSIG_GOOD:
		return DETECT_POWERED_DEVICE;
	case LTC4266_DETECT_RSIG_HIGH:
	case LTC4266_DETECT_OPEN:
		return DETECT_NO_DEVICE;
	default:
		return DETECT_ERROR;
	}
}


/*ltc4266_read_iv: Read the current/voltage of a port
 */
uint32_t ltc4266_read_iv(struct pse_port_device *port, uint8_t iv)
{
	uint8_t lsb_reg;
	int result;
	uint16_t ivbits = 0;
	uint32_t ivval = 0;

	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	if (iv == READ_CURRENT)
		lsb_reg =  LTC4266_IPLSB_REG(port->port_id);
	else if (iv == READ_VOLTAGE)
		lsb_reg =  LTC4266_VPLSB_REG(port->port_id);
	else
		return -EINVAL;

	mutex_lock(&ltc4266->lock);
	result = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_STATPWR);
	if (result < 0)
		goto iv_end;

	/*  LTC4266 IV readings are only valid if the port is powered.*/
	if (result & LTC4266_STATPWR_PG(port->port_id) && (port->power_allocation > 0)) {
		/*Get the LSB....*/
		result = i2c_smbus_read_byte_data(ltc4266->client, lsb_reg);
		if (result < 0)
			goto iv_end;
		ivbits = result & 0xFF;

		/*And then the MSB */
		result = i2c_smbus_read_byte_data(ltc4266->client, lsb_reg + 1);
		if (result < 0)
			goto iv_end;
		ivbits |= ((result & 0xFF) << 8);

		/* We have a "14.5" bit adc on the LTC4266 ports; if either of the two
		 * MSbits are set, we know it's an invalid reading
		 */
		if (ivbits & 0xC000)
			ivbits = 0;
	}
	if (iv == READ_CURRENT) {
		/* Rsense is 0.25 Ohm so 1 LSB = 122.07 uA (see LTC4266 Software
		 * Interface Datasheet).  122.070 uA/LSB, returns uA
		 */
		ivval = (ivbits * 122070)/1000;
	} else {

		/* Rsense is 0.25 Ohm so 1 LSB = 5.835 mV (see LTC4266 Software
		 * Interface Datasheet).  We multiply by factor of 1000 so end result should be
		 * in mV
		 */

		ivval = (ivbits * 5835)/1000;
	}
iv_end:
	mutex_unlock(&ltc4266->lock);
	return ivval;
}

uint8_t ltc4266_port_stat(struct pse_port_device *port)
{
	int result;
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	result = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_STAT_REG(port->port_id));
	if (result < 0)
		return result;
	return (uint8_t) (result & 0xFF);

}

int ltc4266_port_set_ilim(struct pse_port_device *port, int ilim)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	uint8_t ilim_reg_val = ilim_to_reg_val(ilim);

	ltc4266_write_reg(ltc4266, LTC4266_LIM_REG(port->port_id), ilim_to_reg_val(ilim), 0xFF);
	return ilim_from_reg_val(ilim_reg_val);
}

int ltc4266_port_set_icut(struct pse_port_device *port, int icut)
{

	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	/* WARNING: This is saying that you are going to allow the LTC4266 port MOSFET
	 * to source THIS MUCH CURRENT, just FYI. Should you choose to  abuse this
	 * power, you'd best either make sure your FET can take that much current, as
	 * well as your PD, or be prepared to replace both.
	 */
	uint8_t v;

	/* The 1181mA  here is because the function used by the LTC4266 to set the
	 * current limit changes at this point.  Realistically, setting icut higher
	 * than this value is dangerous, setting icut > 850 mA is a violation of the
	 * IEEE802.3at+ spec
	 */

	if (icut > 1181) {
		/* Set the icut range to be able to handle the excessive current */
		v = (uint8_t) ((icut * 100) / 3750);
		dev_warn(&port->dev, "LTC4266 port%d setting current limit to %d mA.\n", port->id, icut);
		v &= LTC4266_CUT_CUT_MSK; /* Make sure the cutrng bit is unset */
	} else {
		v = (uint8_t) ((icut * 100) / 1875);
		v |= LTC4266_CUT_CUTRNG;
	}
	v &= LTC4266_CUT_CUTRNG | LTC4266_CUT_CUT_MSK;
	ltc4266_write_reg(ltc4266, LTC4266_ICUT_REG_HP(port->port_id), v, LTC4266_CUT_CUTRNG|LTC4266_CUT_CUT_MSK);
	return v & LTC4266_CUT_CUT_MSK;
}

int ltc4266_port_get_ilim(struct pse_port_device *port)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	return ilim_from_reg_val(i2c_smbus_read_byte_data(ltc4266->client, LTC4266_LIM_REG(port->port_id)));

}

int ltc4266_port_get_icut(struct pse_port_device *port)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	uint8_t icut_raw = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_ICUT_REG_HP(port->port_id));

	if (icut_raw & LTC4266_CUT_CUTRNG) {
		icut_raw &= LTC4266_CUT_CUT_MSK;
		return (((int)icut_raw) * 1875)/100;
	}
	icut_raw &= LTC4266_CUT_CUT_MSK;
	return (((int)icut_raw) * 3750)/100;
}

int ltc4266_port_highpower(struct pse_port_device *port, uint8_t enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	dev_dbg(&port->dev, "%s:%d\n", __func__, __LINE__);
	dev_dbg(&port->dev, "enable: %d\n", enable);
	if (enable) {
		ltc4266_write_reg(ltc4266, LTC4266_HPEN, LTC4266_HPEN_HPEN(port->port_id), LTC4266_HPEN_HPEN(port->port_id));
		ltc4266_write_reg(ltc4266, LTC4266_HPMD_REG(port->port_id), LTC4266_HPMD_PONGEN, LTC4266_HPMD_PONGEN);
	} else {
		ltc4266_write_reg(ltc4266, LTC4266_HPEN, 0, LTC4266_HPEN_HPEN(port->port_id));
		ltc4266_write_reg(ltc4266, LTC4266_HPMD_REG(port->port_id), 0, LTC4266_HPMD_PONGEN);
	}
	return 0;
}

int ltc4266_port_legacy_enable(struct pse_port_device *port, uint8_t enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	dev_dbg(&port->dev, "%s:%d\n", __func__, __LINE__);
	if (enable) {
		return ltc4266_write_reg(ltc4266, LTC4266_HPMD_REG(port->port_id), LTC4266_HPMD_LEGEN, LTC4266_HPMD_LEGEN);
	} else {
		return ltc4266_write_reg(ltc4266, LTC4266_HPMD_REG(port->port_id), 0, LTC4266_HPMD_LEGEN);
	}
}

int ltc4266_write_reg(struct ltc4266_device *ltc4266, uint8_t reg, uint8_t value, uint8_t mask)
{
	int res;

	mutex_lock(&ltc4266->lock);
	res = i2c_smbus_read_byte_data(ltc4266->client, reg);
	res &= ~mask;
	res |= value & mask;
	res = i2c_smbus_write_byte_data(ltc4266->client, reg, res);
	mutex_unlock(&ltc4266->lock);
	if (res)
		dev_warn(&ltc4266->client->dev, "Failed to write 0x%02X to register 0x%02X\n", value, reg);
	return res;
}
int ltc4266_port_mode(struct pse_port_device *port, uint8_t opmd)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	int result;

	dev_dbg(&port->dev, "%s:%d\n", __func__, __LINE__);
	dev_dbg(&port->dev, "opmd 0x%X\n", opmd);
	if (opmd > OPMD_AUTO)
		dev_warn(&port->dev, "Opmd %d not supported\n", opmd);

	result = ltc4266_write_reg(ltc4266, LTC4266_OPMD_REG, TWO_BIT_WORD_OFFSET(opmd, port->port_id),
			TWO_BIT_WORD_MASK(port->port_id));
	if (result)
		dev_warn(&port->dev, "Failed to write the opmd %d", opmd);
	return result;
}

int ltc4266_port_detect_enable(struct pse_port_device *port, int enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	dev_dbg(&port->dev, "%s(enable=%d)", __func__, enable);
	if (enable == 1) {
		return i2c_smbus_write_byte_data(ltc4266->client, LTC4266_DETPB,  LN(port->port_id));
	} else {
		return ltc4266_write_reg(ltc4266, LTC4266_DETENA, 0, LN(port->port_id));
	}
}

int ltc4266_port_class_enable(struct pse_port_device *port, int enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	dev_dbg(&port->dev, "%s(enable=%d)", __func__, enable);

	if (enable == 1)
		return i2c_smbus_write_byte_data(ltc4266->client, LTC4266_DETPB,  HN(port->port_id));
	else
		return ltc4266_write_reg(ltc4266, LTC4266_DETENA, 0, HN(port->port_id));
}

int ltc4266_port_disconnect_detect_enable(struct pse_port_device *port, int enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	dev_dbg(&port->dev, "%s(enable=%d)", __func__, enable);
	return ltc4266_write_reg(ltc4266, LTC4266_DISENA, ((enable & 0x01) << port->port_id), (1 << port->port_id));
}

int ltc4266_port_powered(struct pse_port_device *port)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	int result = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_STATPWR);

	if (result < 0) {
		dev_err(&port->dev, "Reading LTC4266 STAT POWER FAILED WITH %i\n", result);
		return result;
	}
	dev_dbg(&port->dev, "stat_pwr == 0x02%x\n", result);
	return (result & LTC4266_STATPWR_PG(port->port_id)) || (result & LTC4266_STATPWR_PE(port->port_id));
}

int ltc4266_port_init(struct pse_port_device *port)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);
	int res = 0;

	res |= i2c_smbus_write_byte_data(ltc4266->client, LTC4266_RSTPB, LN(port->port_id));
	res |= ltc4266_port_highpower(port, 1);
	/* Set the sense resistor (0.25 Ohm) value*/
	res |= ltc4266_write_reg(ltc4266, LTC4266_ICUT_REG_HP(port->port_id), LTC4266_CUT_RDIS, LTC4266_CUT_RDIS);
	res |= ltc4266_port_set_ilim(port, 425);
	res |= ltc4266_port_set_icut(port, 375);
	res |= ltc4266_port_disconnect_detect_enable(port, 1);
	res |= ltc4266_port_detect_enable(port, 1);
	res |= ltc4266_port_class_enable(port, 1);
	if (res < 0) {
		dev_err(&port->dev, "Failed to init pse port");
		return res;
	}
	return 0;
}

void ltc4266_enable_interrupts(struct i2c_client *client)
{
	/* Enable the interrupt circuit in the LTC4266 */
	/*Unmask the interrupt sources */
	i2c_smbus_write_byte_data(client, LTC4266_INTMASK, LTC4266_INTERRUPT_SOURCES);
}

int ltc4266_disable_interrupts(struct i2c_client *client)
{
	int r;

	/* Mask out the chip's interrupt */
	r = i2c_smbus_write_byte_data(client, LTC4266_INTMASK, 0x00);
	if (r < 0) {
		dev_err(&client->dev, "Failed to disable interrupts, err %d\n", r);
		return r;
	}
	/* Reset the (SMBus Alert) interrupt pin */
	i2c_smbus_write_byte_data(client, LTC4266_RSTPB, LTC4266_RSTPB_PINCLR);
	return 0;
}


#define IRQ_CHECK_PERIOD 100

void ltc4266_irq_checker(struct timer_list *timer)
{
	struct ltc4266_device *ltc4266 = from_timer(ltc4266, timer, timer);

	mod_timer(&ltc4266->timer, jiffies + msecs_to_jiffies(IRQ_CHECK_PERIOD));
	if (gpio_get_value(LTC4266_INTERRUPT_PIN) == 0) {
		schedule_work(&ltc4266->work_struct);
	}
}
static void plat_enable_pse_interrupt(struct ltc4266_device *ltc4266)
{
	timer_setup(&ltc4266->timer, ltc4266_irq_checker, 0);
	mod_timer(&ltc4266->timer,  jiffies + msecs_to_jiffies(IRQ_CHECK_PERIOD));
}

static void plat_disable_pse_interrupt(struct ltc4266_device *ltc4266)
{
	del_timer(&ltc4266->timer);
}

int ltc4266_port_power(struct pse_port_device *port, int enable)
{
	struct ltc4266_device *ltc4266 = pse_port_get_clientdata(port);

	if (!enable) {
		i2c_smbus_write_byte_data(ltc4266->client, LTC4266_PWRPB, HN(port->port_id));
		ltc4266_port_init(port);
	} else {
		i2c_smbus_write_byte_data(ltc4266->client, LTC4266_PWRPB, LN(port->port_id));
	}
	return  0;
}

extern uint32_t pse_port_power_table[]; /* in pse-sysfs.c */
void ltc4266_read_events(struct work_struct *work)
{
	int detevn = 0, supevn = 0, pwrevn = 0;
	int fltevn = 0, tsevn = 0;
	int intstat;
	struct pse_port_device *p;
	struct ltc4266_device *ltc4266 = container_of(work, struct ltc4266_device, work_struct);
	int i = 0;

	/* While it should never happen that we don't have a valid i2c client here,
	 * it may be possible when the module is removed.
	 */
	if (!ltc4266)
		return;

	ltc4266_disable_interrupts(ltc4266->client);

	intstat = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_INTSTAT);
	if (intstat < 0) {
		pr_alert("Bus failure during attempt to read INTSTAT\n");
		goto done;
	}
	if (!intstat) {
		goto done;
	}
	dev_dbg(&ltc4266->client->dev, "intstat = 0x%02x", (uint8_t) (intstat & 0xFF));
	if (intstat & LTC4266_INT_SUPPLY) {
		supevn = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_SUPEVN_COR);
		if (supevn < 0) {
			dev_err(&ltc4266->client->dev, "Bus error reading SUPPLY Fault register");
			goto done;
		}
		if (supevn) {
			dev_err(&ltc4266->client->dev, "SUPPLY_FAULT, 0x%02X", supevn);
			goto done;
		}
	}

	if (intstat & (LTC4266_INT_DIS | LTC4266_INT_TCUT)) {
		fltevn = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_FLTEVN_COR);
		if (fltevn < 0) {
			dev_err(&ltc4266->client->dev, "Failed to read fltevn\n");
			goto done;
		}
		fltevn &= 0xFF;
	}

	if (intstat & (LTC4266_INT_TSTART | LTC4266_INT_TCUT)) {
		tsevn = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_TSEVN_COR);
		if (tsevn < 0) {
			dev_err(&ltc4266->client->dev, "Failed to read tsevn\n");
			goto done;
		}
		tsevn &= 0xFF;
	}

	if (intstat & (LTC4266_INT_PWRGD | LTC4266_INT_PWRENA)) {
		pwrevn = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_PWREVN_COR);
		if (pwrevn < 0) {
			dev_err(&ltc4266->client->dev, "Failed to read pwrevn\n");
			goto done;
		}
		pwrevn &= 0xFF;
	}

	if (intstat & (LTC4266_INT_DET | LTC4266_INT_CLASS)) {
		detevn = i2c_smbus_read_byte_data(ltc4266->client, LTC4266_DETEVN_COR);
		if (detevn < 0) {
			dev_err(&ltc4266->client->dev, "Failed to read pwrevn\n");
			goto done;
		}
		detevn &= 0xFF;
	}

	for (i = 0; i < LTC4266_NPORTS; i++) {
		p = ltc4266->pse_ports[i];
		if ((tsevn | fltevn) & (HN(i) | LN(i))) {
			queue_work(p->wq, &p->disconnect);
			/* If a port has disconnected, we don't care about (possibly old)
			 * detection/classification events.  Even if they are "new", we'll see them
			 * again (without a disconnect event) at the next detection/classification
			 * cycle
			 */
			continue;
		}
		if (detevn & HN(i)){
			mutex_lock(&p->lock);
			p->last_detect_result = p->ops->last_detection(&p->dev);
			p->last_class_result = p->ops->last_classification(&p->dev);
			p->power_request = pse_port_power_table[p->last_class_result];
			queue_work(p->wq, &p->request_power);
			mutex_unlock(&p->lock);
		}
	}
done:
	ltc4266_enable_interrupts(ltc4266->client);

}


static int ltc4266_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	int pc;
	int res;
	uint8_t id_reg;
	struct ltc4266_device *ltc4266_data;
	struct pse_port_device *port;

	dev_dbg(&client->dev, "%s\n", __func__);
	dev_info(&client->dev, "Client data passed at address %p", client);
	dev_info(&client->dev, "I2C Address %x\n", client->addr);
	dev_info(&client->dev, "name: %s\n", client->name);

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C))
		return -ENODEV;

	/* The id_reg value here is not what was specified in the datasheet, nor
	 * the include file. Lets make sure we've got something at address 0x2F that
	 * happens to have a register at 0x1B that returns 0x64
	 */
	id_reg = i2c_smbus_read_byte_data(client, LTC4266_ID);
	if (id_reg != 0x64) {
		dev_warn(&client->dev, "Expected an ID of 0x64, saw 0x%02X\n", id_reg);
		return -ENODEV;
	}

	/* Since we're reasonably sure we've got the right chip, lets register the
	 * ports with the pse class
	 */

	ltc4266_data = devm_kzalloc(&client->dev, sizeof(struct ltc4266_device), GFP_KERNEL);
	if (!ltc4266_data)
		return -ENOMEM;

	mutex_init(&ltc4266_data->lock);
	i2c_set_clientdata(client, ltc4266_data);

	/* We are sure enough that this is our device, we're willing to save the i2c_client pointer */
	ltc4266_data->client = client;

	i2c_smbus_write_byte_data(client, LTC4266_RSTPB,  LTC4266_RSTPB_INTCLR|LTC4266_RSTPB_RSTALL);

	/* LTC4266 requires approximately 10 ms after reset to be stable; if it
	 * isn't, then there is typically an undervoltage lockout/something pretty bad
	 * going on. We give it 50 ms here so we don't need to poll the chip and use I2C bandwidth
	 */

	msleep(50);

	/* Let's make sure the chip came out of reset (if not, the chip is probably
	* either (no longer?) present, in thermal shutdown, or watchdogged....either
	* way, there's nothing we can do in software to fix it
	*/
	id_reg = i2c_smbus_read_byte_data(client, LTC4266_ID);
	if (id_reg != 0x64) {
		dev_err(&client->dev, "Expected an ID of 0x64, saw 0x%02X\n", id_reg);
		return -ENODEV;
	}

	/* Some interrupts are enabled after reset, disable those here */
	ltc4266_disable_interrupts(client);

	/*Set all the ports into Semi OPMD (automatically run detect & classify and report, wait for command to power */
	i2c_smbus_write_byte_data(client, LTC4266_OPMD_REG, LTC4266_OPMD_SEMI);

	/* After reset, the LTC4266 will  interrupt with a (single) supply fault.
	 * It's not a bit deal, but it looks pretty nasty in the logs, so we clear it
	 * here and discard the result
	 */
	i2c_smbus_read_byte_data(client, LTC4266_SUPEVN_COR);

	for (pc = 0; pc < LTC4266_NPORTS; pc++) {
		port = devm_kzalloc(&client->dev, sizeof(struct pse_port_device), GFP_KERNEL);
		if (!port) {
			res = -ENOMEM;
			pc--; /*Back out the index by one to not segfault while unregistering */
			goto class_error;
		}
		port->ops = &ltc4266_port_ops;
		port->owner = THIS_MODULE;
		port->dev.parent = &client->dev;
		strlcpy(port->name,ltc4266_driver.driver.name,  PSE_PORT_NAME_SIZE);
		pse_port_set_clientdata(port, (void *) ltc4266_data);

		port->port_id = pc;
		ltc4266_data->pse_ports[pc] = port;

		pse_port_device_register(port);
		dev_info(&client->dev, "ltc4266_probe has registered port %d on the LTC4266 as port%d\n", port->port_id, port->id);
		dev_info(&port->dev, "Has been initialized");
	}

	/*Create the workqueue used for the event processing */
	INIT_WORK(&ltc4266_data->work_struct, (void (*)(struct work_struct *)) ltc4266_read_events);

	/*Enable the eventing pin on the SoC now that we've requested an IRQ */
	plat_enable_pse_interrupt(ltc4266_data);

	/*Enable LTC4266 interrupt circuitry */
	ltc4266_enable_interrupts(client);
	return 0;

	pr_alert("LTC4266 failed to request IRQ\n");
class_error:
	for (; pc >= 0; pc--)
		pse_port_device_unregister(ltc4266_data->pse_ports[pc]);
	return res;
}

static int ltc4266_remove(struct i2c_client *client)
{
	int pc;
	struct ltc4266_device *ltc4266 = i2c_get_clientdata(client);

	/*Disable the LTC4266's interrupt mechanism */
	ltc4266_disable_interrupts(client);

	plat_disable_pse_interrupt(ltc4266);

	/*Put the LTC4266 into reset */
	i2c_smbus_write_byte_data(client, LTC4266_RSTPB,  LTC4266_RSTPB_RSTALL);

	for (pc = 0; pc < LTC4266_NPORTS; pc++) {
		pse_port_device_unregister(ltc4266->pse_ports[pc]);
		devm_kfree(&client->dev, ltc4266->pse_ports[pc]);
	}

	devm_kfree(&client->dev, ltc4266);
	return 0;
}
static const struct i2c_device_id ltc4266_id[] = {
	{ "ltc4266", 0},
	{ }
};
MODULE_DEVICE_TABLE(i2c, ltc4266_id);

static const unsigned short normal_i2c[] = {0x2F, I2C_CLIENT_END};

static const struct of_device_id ltc4266_of_match[] = {
	{ .compatible = "ltc,ltc4266" },
	{ }
};

MODULE_DEVICE_TABLE(of, ltc4266_of_match);

static struct i2c_driver ltc4266_driver = {

	.driver		= {
		.name	= "ltc4266",
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(ltc4266_of_match),
	},
	.probe		= ltc4266_probe,
	.remove		= ltc4266_remove,
	.id_table	= ltc4266_id,
	.address_list = normal_i2c,
};

module_i2c_driver(ltc4266_driver);


MODULE_AUTHOR("Kyle Swenson <kswenson@cradlepoint.com>");
MODULE_DESCRIPTION("LTC4266 I2C Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
